/*
 *  UUStat for Waffle-UUCP.
 *  Written by Jochen Erwied <mack@wes.open.de>
 *
 *  Compiler: Borland C++ 3.1. Should work on any other compiler with
 *  only small changes.
 *  Tested with: GNU-C V2.2.2 (no timezone for Newsheader!)
 *
 *  Version 1.0  - 10/26/92
 * 			Initial day of programming
 *
 *  Version 1.01 - 10/26/92
 * 			Bugfix: If a uucico session was terminated, but never correctly
 * 				started, uustat would add the duration from the last valid
 * 				call until now.
 *
 *  Version 1.02 - 10/28/92
 * 			Produce a news-header
 * 			Scan WAFFLE-Static for system & node!
 * 			Use strftime() to compute correct date & subject
 *
 *  Version 1.021 - 01/11/92
 * 			If called on 1st of month, assume this to be a statistic
 * 			for the last month, and produce an according header.
 * 			Fix to produce correct program-name (GNU-C)
 *
 *  Version 1.022 - 11/11/92
 * 			Bugreport by <peters@bsc.guild.org>:
 * 			For sites receiving over 21MB of data, average CPS is
 * 			calculated wrong. Fixed.
 *
 *  Version 1.03 - 19/11/92
 * 			Added -1 switch - subtract one from current month for
 * 			generating statistics for the last month (-n only)
 * 			Added -s switch to specify an alternate subject for
 * 			Newsarticles
 * 			These options supersede some things changes in V1.021
 * 			Added an #ifdef for __MSC__ as reported by <peters@bsc>
 *
 *  Purpose:
 *
 * 	  The program scans uucico's logfile and produces a statistic.
 * 	  UUCICO must be called with '-x3' to produce enough information
 * 	  to analyze.
 */

/* Program version */
#define VERSION "1.03"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <time.h>

/* To get MAXLONG */
#ifdef __MSC__
/* Microsoft-C */
#include <limits.h>
#define MAXLONG LONG_MAX
#else
/* Borland-C, GNU-C */
#include <values.h>
#endif

/* Constants
 * FULLDAY - 100ths of seconds per day. Used for computing times
 */
#define FULLDAY 24L*60L*60L*100L
#define BUFLEN 256

/* Defines for sorting */
#define BYTES 1
#define HOURS 2
#define CPS 4
#define FILES 8

#define ASCENDING 16
#define RECEIVED 32
#define SENT 64
#define TOTAL 128

/* Prototypes for external Routines */
extern int getopt(int argc, char **argv, char *opts);
extern char *optarg;
extern char *getconfig(char *option);

/* Structure to hold information about the various sites */
struct hostlist {
	char name[9];
	long sent_bytes, received_bytes;
	long sent_time, received_time;
	long sent_files, received_files;
	long connection_time;
	int printed;
	struct hostlist *next;
};

/* Global top-of-list */
struct hostlist *top = NULL;

/* Local functions and their prototypes */
void outhelp(void);
void outheader(char *newsgroup);
void process(char *filename);
long maketime(char *buffer);
struct hostlist *addhost(char *name);
long comptime(long start,long end);
long getbytes(char *buffer);
void printstat(void);
char *printtime(long timer);
int compare(long *max, struct hostlist *help);
long average(long a_time,long a_bytes);

/* How should the list be sorted by default? */
int sortopt = BYTES|TOTAL;
int submonth = 0;
char *subject=NULL;

/* Not the main job, but the main routine. Just a bit of option
 * parsing, and searching for the correct file to scan
 */
int
main(int argc, char **argv)
{
	char uufile[128];
	char *newsgroup = NULL;
	char waffle[128], *help;
	char c;
	int cnt;

	strcpy(uufile, "");

	if ( ( ( help = strrchr (argv[0], '\\')) != NULL ) || ( ( help = strrchr(argv[0], '/') ) != NULL ) )
		argv[0] = help + 1;
	if ( ( help = strrchr(argv[0], '.') ) != NULL )
		*help = '\0';

	while ((c = getopt(argc,argv,"1n:F:k:h:c:s:f:?")) != EOF)
	{
		switch (c)
		{
			case 'F':
					strcpy(uufile, optarg);
					break;
			case 'n':
					newsgroup = optarg;
					break;
			case 's':
					subject = optarg;
					break;
			case '1':
					submonth--;
					break;
			case 'k':
					sortopt = BYTES;
					c = 'k';
					goto SetSort;
			case 'h':
					sortopt = HOURS;
					c = 'h';
					goto SetSort;
			case 'c':
					sortopt = CPS;
					c = 'c';
					goto SetSort;
			case 'f':
					sortopt = FILES;
					c = 'f';
SetSort:
					if( strchr ("rst", *optarg) == NULL )
						printf("Warning: illegal option for '-%c' (%c) - ignored\n", c, *optarg);
					else
					{
						if ( *optarg == 'r' ) sortopt |= RECEIVED;
						if ( *optarg == 's' ) sortopt |= SENT;
						if ( *optarg == 't' ) sortopt |= TOTAL;
						if ( *(optarg+1) == '+' ) sortopt |= ASCENDING;
					}
					break;
			case '?':
					outhelp();
					exit(0);
					break;
		}
	}

	if (!strlen(uufile))
	{
		if ((help = getenv("WAFFLE")) == NULL)
		{
			printf("Fatal error: WAFFLE environment-variable not specified.\n"
					 "             Use '-F<filename>' to specify an alternate logfile.\n");
		}
		else
		{
			strcpy(waffle, help);
			help = waffle + strlen(waffle) - 1;
			cnt = 0;
			while ( (help != waffle) && (cnt != 2) )
			{
				if (*help == '\\' || *help == '/')
					cnt++;
				help--;
			}
			if ( cnt == 2 )
			{
				*++help = '\0';
				sprintf(uufile,"%s/admin/uucico", waffle);
			}
			else
				printf("Fatal error: WAFFLE environment-variable has illegal format.\n");
		}
	}

	if ( strlen(uufile) ) {
		process ( uufile );
		if(newsgroup != NULL)
			outheader(newsgroup);
		printstat ();
	}

	return 0;
}

/* Print out help text */
void
outhelp(void)
{
	printf("UUStat for Waffle  -  written by Jochen Erwied <mack@wes.open.de>\n"
			 "                      Version %s, last compiled: %s %s\n"
			 "Options:\n"
			 "\t-c[rst]{+}    sort list descending on CPS-rate (Received, Sent, Total)\n"
			 "\t-f[rst]{+}    sort list descending on files\n"
			 "\t-h[rst]{+}    sort list descending on hours\n"
			 "\t-k[rst]{+}    sort list descending on K-bytes\n"
			 "\t              default: -kt, append '+' for ascending sort\n"
			 "\t-F<filename>  specify an alternate logfile\n\n"
			 "\t-n<newsgroup> produce an output to be fed to rnews\n"
			 "\t-1            subtract one from current month (cummulative)\n"
			 "\t-s<subject>   specify subject for newsarticles\n"
			 "\n"
			 "\t-?            print this help\n",
			 VERSION, __DATE__, __TIME__);
}

/* Produces a news-header */
void
outheader(char *newsgroup)
{
	time_t timer;
	struct tm *timmy;
	char domain[128], *help;
	char localsubject[256];

	if ((help = getconfig("node")) == NULL)
		strcpy(domain, "UnKnown.SubDomain.Domain");
	else
		strcpy(domain, help);

	tzset();
	time(&timer);
	timmy = gmtime(&timer);

	/* Adjust month according to date - subtract a number of months from
		the current date, according to the flag submonth
	 */

	while (submonth < 0)
	{
		if(timmy->tm_mon == 1)
		{
			timmy->tm_mon=12;
			timmy->tm_year--;
		}
		else
			timmy->tm_mon--;
		submonth++;
	}

	/* Produces a default subject to be displayed, and if specified by '-s'
		an alternative */

	strftime(localsubject, (size_t) 256, "Monthly newsstatistic for %B %Y", timmy);
	if(subject != NULL)
		strftime(localsubject, (size_t) 256, subject, timmy);

	printf("Path: root\n"
			 "From: root@%s\n"
			 "Subject: %s\n"
			 "Newsgroups: %s\n"
			 "Message-ID: <%ld@%s>\n"
			 "X-Version: UUStat V%s for Waffle by Jochen Erwied <mack@wes.open.de>\n",
			 domain,
			 localsubject,
			 newsgroup,
			 (long) timmy, domain, VERSION);

#ifdef __GNUC__
	/* No TimeZone flag for strftime() by default! */
	strftime(domain, 128, "%a, %d %b %y %H:%M:%S", timmy);
#else
	strftime(domain, 128, "%a, %d %b %y %H:%M:%S %Z", timmy);
#endif
	printf("Date: %s\n\n", domain);
}

/* Scan the Logfile */
void
process(char *name)
{
	FILE *fp = fopen(name, "r");
	char *help;
	char buffer[BUFLEN];
	int sendflag;
	struct hostlist *currhost;
	long starttime = -1, endtime;
	long startfile, endfile;
	long bytes, seconds;
	long totbytes, totseconds;

	if ( fp == NULL )
	{
		printf("Fatal error: Could not open '%s'.\n", name);
		return;
	}

	while ( fgets(buffer, BUFLEN, fp) != NULL )
	{
		if ( (help = strstr(buffer, "Talking")) != NULL ) {
			help = help + 12;
			*(strchr(help, ',')) = '\0';
			currhost = addhost (help);
			starttime = maketime(buffer);
			totbytes = 0L;
			continue;
		}
		if ( (strstr(buffer, "packets") != NULL) ||
			  (strstr(buffer, "Terminated") != NULL) )
		{
			if (starttime != -1L)
			{
				endtime = maketime(buffer);
				if (endtime < starttime)
					endtime+= FULLDAY;
				totseconds = comptime(starttime, endtime);
				starttime = -1L;
				currhost->connection_time += totseconds;
			}
			continue;
		}
		if ( strstr(buffer, "sending") != NULL || strstr(buffer, "receiving") != NULL)
		{
			startfile = maketime(buffer);
			continue;
		}
		if ( strstr(buffer, " sent") != NULL || strstr(buffer, "received") != NULL)
		{
			sendflag = ( strstr(buffer, "sent") != NULL );
			endfile = maketime(buffer);
			bytes = getbytes(buffer);
			totbytes += bytes;
			seconds = comptime(startfile, endfile);

			if (sendflag)
			{
				currhost->sent_time += seconds;
				currhost->sent_bytes += bytes;
				currhost->sent_files++;
			}
			else
			{
				currhost->received_time += seconds;
				currhost->received_bytes += bytes;
				currhost->received_files++;
			}
			continue;
		}
	}

	fclose(fp);
}

/* This routine converts the string '23:30:00.88' into a 'global' time,
 * measured in 100ths of seconds since midnight
 */
long
maketime(char *buffer)
{
	long hour, minute, second, hundred;

	if ( sscanf(buffer, "%ld:%ld:%ld.%ld", &hour, &minute, &second, &hundred) != 4)
	{
		printf("Warning: Incorrect time!?\n");
		return 0L;
	}
	else
		return hundred + 100L * (second + 60L * (minute + 60L * hour));
}

/* Scan through the list of available hosts, return a pointer to the
 * correct host, if found, else append to list and return this pointer
 */
struct hostlist *
addhost(char *name)
{
	register struct hostlist *help=top, *last=NULL;

	char locname[128];
	strncpy(locname, name, 8);
	if ( strchr(locname, '.') != NULL )
		*(strchr(locname, '.')) = '\0';

	while( help!=NULL )
	{
		if(!strcmp(help->name, locname))
			return help;
		else {
			last = help;
			help = help->next;
		}
	}

	help = malloc( sizeof(struct hostlist) );

	if (help == NULL) {
		printf("Fatal error: Couldn't allocate memory for host '%s'!\n", name);
		exit(0);
	}

	if (last == NULL) {
		top = help;
		top->next = NULL;
	} else
		last->next = help;

	strcpy(help->name, locname);
	help->sent_files = 0L;
	help->received_files = 0L;
	help->sent_bytes = 0L;
	help->received_bytes = 0L;
	help->sent_time = 0L;
	help->received_time = 0L;
	help->connection_time = 0L;
	help->printed = 0;
	help->next = NULL;

	return help;
}

/* Compare to times evaluated by maketime with testing for polls from
 * 23:59:59 to 00:05:00
 */
long
comptime(long start,long end)
{
	if (end < start)
		end += FULLDAY;
	return end - start;
}

/* Extract the correct filesize of the last received/sent file from the
 * line pointed to by buffer.
 */
long
getbytes(char *buffer)
{
	char *help;

	help = strrchr(buffer, '(');
	if ( help == NULL )
		return 0L;
	else
		return atol(help+1);
}

/* Print out statistic.
 * The format of this table was stolen from uutraf, and a bit extended
 * to suit my own needs.
 */
void
printstat(void)
{
	struct hostlist *print, *help;
	long max;
	char sent[7], received[7], total[7];
	int sites = 0, doprint = 1;
	long filessent = 0, filesreceived = 0, bytessent = 0, bytesreceived = 0;
	long timesent = 0, timereceived = 0, timetotal = 0;

	if ( top == NULL )
		return;

	printf(" Remote  ------ K-bytes ----- ------ Hours ------- -- Avg. CPS -- -- Files --\n"
			 "  host    Recv   Sent   Total  Recv   Sent   Total Recv Sent Totl  Recv  Sent\n"
			 "-------- ------ ------ ------ ------ ------ ------ ---- ---- ---- ----- -----\n");

	while ( doprint )
	{
		help = top;
		max =  (sortopt & ASCENDING) ? MAXLONG : -MAXLONG;
		print = NULL;
		doprint = 0;
		while ( help != NULL )
		{
			if ( !help->printed )
			{
				if ( compare (&max, help) )
					print = help;
			}
			help = help->next;
		}
		if ( print != NULL )
		{
			printf("%-8s %6ld %6ld %6ld %6s %6s %6s %4ld %4ld %4ld %5ld %5ld\n",
				print->name,
				print->received_bytes / 1024L,
				print->sent_bytes / 1024L,
				(print->received_bytes + print->sent_bytes) / 1024L,
				strcpy (received, printtime(print->received_time)),
				strcpy (sent, printtime(print->sent_time)),
				strcpy (total, printtime(print->connection_time)),
				average(print->received_time, print->received_bytes),
				average(print->sent_time, print->sent_bytes),
				average(print->connection_time, print->received_bytes + print->sent_bytes),
				print->received_files,
				print->sent_files);

			sites++;
			bytessent += print->sent_bytes;
			bytesreceived += print->received_bytes;
			filessent += print->sent_files;
			filesreceived += print->received_files;
			timesent += print->sent_time;
			timereceived += print->received_time;
			timetotal += print->connection_time;
			print->printed = 1;
			doprint = 1;
		}
	}

	while ( top != NULL )
	{
		help = top;
		top = top->next;
		free (help);
	}

	if (sites)
	{
		printf("\n\t\t------------ Summary --------------"
				 "\n\t\tTotal UUCP-Sites       : %7d"
				 "\n\t\tTotal K-bytes received : %7ld"
				 "\n\t\tTotal K-bytes sent     : %7ld"
				 "\n\t\tTotal K-bytes          : %7ld"
				 "\n\t\tTotal files received   : %7ld"
				 "\n\t\tTotal files sent       : %7ld"
				 "\n\t\tTotal files            : %7ld"
				 "\n\t\tTotal hours received   : %10s"
				 "\n\t\tTotal hours sent       : %10s"
				 "\n\t\tTotal hours            : %10s\n",
			sites,
			bytesreceived / 1024L,
			bytessent / 1024L,
			(bytesreceived + bytessent) / 1024L,
			filesreceived,
			filessent,
			filesreceived+filessent,
			strcpy ( received, printtime(timereceived) ),
			strcpy ( sent, printtime(timesent) ),
			strcpy ( total, printtime(timetotal) ));
	}
}

/* Convert an amount of 100ths of a second to a printable format in hours.
 */
char *
printtime(long timer)
{
	static char buffer[7];
	long hours, minutes;

	minutes = timer / (100L * 60L);
	hours = minutes / 60;
	minutes %= 60L;

	sprintf(buffer, "%3ld:%2ld", hours, minutes);
	if (buffer[4] == ' ')
		buffer[4] = '0';

	return buffer;
}

/* Compare a value with a given maximum, using the currently active sorting
 * order.
 * This could be done more intelligent, but it works...
 */
int
compare(long *max, struct hostlist *help)
{
	long cmp = 0L;

	if ( sortopt & BYTES )
	{
		if ( sortopt & SENT ) cmp = help->sent_bytes;
		if ( sortopt & RECEIVED ) cmp = help->received_bytes;
		if ( sortopt & TOTAL ) cmp = help->received_bytes + help->sent_bytes;
	}
	if ( sortopt & HOURS )
	{
		if ( sortopt & SENT ) cmp = help->sent_time;
		if ( sortopt & RECEIVED ) cmp = help->received_time;
		if ( sortopt & TOTAL ) cmp = help->connection_time;
	}
	if ( sortopt & CPS )
	{
		if ( sortopt & SENT ) cmp = ( help->sent_time == 0L ? 0L : (help->sent_bytes * 100L ) / help->sent_time ) ;
		if ( sortopt & RECEIVED ) cmp = ( help->received_time == 0L ? 0L : (help->received_bytes * 100L ) / help->received_time );
		if ( sortopt & TOTAL ) cmp = ((help->sent_bytes + help->received_bytes) * 100L ) / help->connection_time;
	}
	if ( sortopt & FILES )
	{
		if ( sortopt & SENT ) cmp = help->sent_files;
		if ( sortopt & RECEIVED ) cmp = help->received_files;
		if ( sortopt & TOTAL ) cmp = help->sent_files + help->received_files;
	}

	if ( (sortopt & ASCENDING) ? (cmp < *max) : (cmp > *max) )
	{
		*max = cmp;
		return 1;
	}
	else
		return 0;
}

/* Compute average CPS for a given number of bytes and a given time in
 * hundreths of a second.
 * Added in Version 1.022
 * Statistics for more than 21.5MB of data resulted in wrong results
 * since 21.5*10^6 * 100 is greater than MAXLONG,
 * thus a negative number :-(
 */
long
average(long a_time,long a_bytes)
{
	long h1, h2;

	if (a_time == 0L)
		return 0L;

	h1 = (a_bytes / a_time) * 100L ;
	h2 = ((a_bytes % a_time) * 100L) / a_time ;

	return h1+h2;
}